/*********************************************************************************
 *
 *  The C++ binding for LDB
 *  Copyright (c) 2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 02/07/03
 *
 ********************************************************************************/

#ifndef _Included_LDB
#define _Included_LDB

#include <stream.h>
#include <string>

typedef std::string String;


const char* cform ( const char* fmt, ... );


class ServerError {
public:
    String message;
    ServerError ( String msg ) { message=msg; };
    friend ostream& operator<< ( ostream& o, const ServerError &s ) { return o << s.message; };
};


class LDBobject {
private:
  int ID;
  void operation ( short opr ) const;
  void flush ();
public:
  LDBobject ( int id ) { ID=id; };
  LDBobject ();			                        // create a new object at the server
  ~LDBobject () { flush(); };	                        // remove the object from both client and server
  String ref () const { return cform("(:%d)",ID); };    // for embedding the object in a query
  int getInt () const;		                        // if the object is an int, return its value
  String getString () const;		                // if the object is a string, return its value
  float getFloat () const;		                // if the object is a float, return its value
  int size () const;			                // if the object is a collection, return its size
  LDBobject nth ( int index ) const;	                // if the object is a collection, return the nth element
  String toString () const;	                        // pretty-print the value of the object into a string
  String getType () const;	                        // return the type of the object (eg, "set(Employee)")
  friend ostream& operator<< ( ostream& o, const LDBobject &obj ) { return o << obj.toString(); };
};


class LDB {
  bool opened;
  bool in_transaction;
public:
  LDB () { opened=false; in_transaction=false; };
  void close ();                                        // logout
  ~LDB () { close(); };
  void open ( String host, String username,	        // login to host using username and
	      String password, String schema );		//    password and import schema
  void open ( String schema ) { open("127.0.0.1","admin","pass",schema); };
  LDBobject query ( String query );			// evaluate query
  void statement ( String stmt );
  void update ( String dest, String query );		// do dest := query
  void insert ( String dest, String query );		// do dest += query
  void remove ( String dest, String query );		// do dest -= query
  void create_index ( String index_name, String class_name,
		      String attribute );		// create a new named index
  void remove_index ( String index_name );		// drop the index
  void change_password ( String oldp, String newp );	// change user password
  String tcl_schema ();                                 // return the schema in TCL form
  void begin_transaction ();
  void commit_transaction ();
  void abort_transaction ();
  void process_definition ( const char* definition );
};

#endif
