/*********************************************************************************
 *
 *  Populating the School database
 *  Copyright (c) 2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 4/14/03
 *
 ********************************************************************************/


class Populate {
    public static void main ( String args[] ) throws Exception {
	LDB ldb = new LDB();
	try {
	    ldb.open("School");
	    ldb.begin_transaction();

	    // populate Persons
	    LDBobject p1 = ldb.query("persistent Person( ssn: 1234, name: 'Smith')");
	    LDBobject p2 = ldb.query("persistent Person( ssn: 5678, name: 'Jones',"
				     + "address: struct(street: 'Park Av', state: 'TX',"
				     + "city: 'Dallas', zip: 123))");

	    // populate departments
	    LDBobject cse = ldb.query("persistent Department( name: 'CSE', dno: 1 )");
	    LDBobject ee = ldb.query("persistent Department( name: 'EE', dno: 2 )");

	    // populate instructors
	    LDBobject i1 = ldb.query("persistent Instructor( name: 'John', rank: 'Professor', dept: " + cse.ref() + ","
				     + "degrees: set('BS','MS','PhD'), salary: 100000, ssn: 3456,"
				     + "address: struct(street: 'Ross St.', state: 'TX', city: 'Arlington', zip: 124))");
	    LDBobject i2 = ldb.query("persistent Instructor( name: 'Mary', rank: 'Instructor', dept: " + cse.ref() + ","
				     + "degrees: set('BS','MS'), salary: 80000, ssn: 6347,"
				     + "address: struct(street: '4th St', state: 'TX', city: 'Dallas', zip: 123))");
	    LDBobject i3 = ldb.query("persistent Instructor( name: 'Helen', rank: 'Professor', dept: " + ee.ref() + ","
				     + "degrees: set('BS','MS','PhD'),salary: 100000, ssn: 2038,"
				     + "address: struct(street: 'Main St', state: 'TX', city: 'Fort Worth', zip: 125))");
	    ldb.update(cse.ref()+".head",i1.ref());
	    ldb.update(ee.ref()+".head",i3.ref());

	    // populate courses
	    LDBobject c1 = ldb.query("persistent Course( code: 'cse5330', name: 'DB I', taught_by: "
				     + i1.ref() + ", offered_by: " + cse.ref() + " )");
	    LDBobject c2 = ldb.query("persistent Course( code: 'cse5331', name: 'DB II', taught_by: "
				     + i2.ref() + ", offered_by: " + cse.ref()
				     + ",has_prerequisites: set(" + c1.ref() + "))");

	    ldb.commit_transaction();
	    ldb.close();
	} catch (ServerError m) {
	    System.out.print(m);
	    ldb.close();
	}
    }
}
