
%{
#include "external.h"
#include "xml.yy.cc"

static Expr xml_ast;

Expr L ( char* s, list<Expr>* r ) { return function(variable(s),r); };
Expr L ( char* s, Expr x1, list<Expr>* r ) { return function(variable(s),Cons(x1,r)); };
Expr F ( char* s, Expr x1, Expr x2 ) { return function(variable(s),Cons(x1,Cons(x2,Nil))); };
Expr F ( char* s, Expr x1, Expr x2, Expr x3 ) { return function(variable(s),Cons(x1,Cons(x2,Cons(x3,Nil)))); };


Expr element ( Expr name, list<Expr>* attributes, list<Expr>* content ) {
  const Expr elem = variable("element");
  const Expr attr = variable("attributes");
  const Expr cont = variable("content");
  return function(elem,Cons(name,Cons(function(attr,attributes),Cons(function(cont,content),Nil))));
};


Expr bind ( Expr a, Expr b ) {
  const Expr bd = variable("bind");
  return function(bd,Cons(a,Cons(b,Nil)));
};


int yyerror ( const char* text ) {
  cerr << "*** " << text << " near the token '" << yytext << "'\n";
  exit(-1);
};

int yylex();

%}

%start program

%union {
        String        ST;
        Expr	      EType;
        list<Expr>*   ElistType;
}

%token BEGINT END ENDL ENDR ID STRINGT EQ TOKEN DOCTYPE

%type <ST> token tokens
%type <EType> name string element program
%type <ElistType> attributes attrs content

%%
program: content			{ $$ = L("content",$1->reverse()); xml_ast = $$; }
       ;

name: ID                                { $$ = variable(new string(yytext)); }
    ;

token: TOKEN                            { $$ = new string(yytext); }
    ;

string: STRINGT				{ yytext[strlen(yytext)-1]='\0';
                                          $$ = estring(new string(yytext+1)); }
      ;

element: BEGINT name attrs END content ENDL name END
	       { if (!$2->eq($7))
	            yyerror((string("Unmatched tag names: ")+*$2->name()+*$7->name()).content());
		 $$ = element($2,$3,$5->reverse());
	       };
       | BEGINT name attrs ENDR		{ $$ = element($2,$3,Nil); }
       | DOCTYPE name name string END	{ $$ = F("dtd",$2,$4); }
       ;

attrs: attributes			{ $$ = $1->reverse(); }
     |					{ $$ = Nil; }
     ;

attributes: name EQ string		{ $$ = Cons(bind($1,$3),Nil); }
	  | attributes name EQ string	{ $$ = Cons(bind($2,$4),$1); }
	  ;

content: element			{ $$ = Cons($1,Nil); }
       | tokens				{ $$ = Cons(estring($1),Nil); }
       | content element		{ $$ = Cons($2,$1); }
       | content tokens			{ $$ = Cons(estring($2),$1); }
       ;

tokens: token				{ $$ = $1; }
      | tokens token			{ $$ = new string(*$1 + *$2); }
      ;

%%


Expr download ( const char* file ) {
  if (file)
     yyin = fopen(file,"r");
  yyparse();
  return xml_ast;
};
