#--------------------------------------------------------------------------------
#
# RPC functions for the OQL interpreter
# Programmer: Leonidas Fegaras
# Creation: 02/12/03
# Copyright (c) 2003 by Leonidas Fegaras, the University of Texas at
#     Arlington. All rights reserved.
#
#--------------------------------------------------------------------------------


# the default server is the local server
set server "127.0.0.1"


# The port number of the server.
set port_number 9737


set rpc_tags {flush getInt getString getFloat size nth toString
              getType open close query update insert remove createi
              dropi change tclschema begin commit abort}


proc command { tagname } {
    global rpc_tags socket
    send_short [lsearch -exact $rpc_tags $tagname]
}


proc send_short { n } {
    global socket
    puts -nonewline $socket [binary format S $n]
    flush $socket
}


proc send_integer { n } {
    global socket
    puts -nonewline $socket [binary format I $n]
    flush $socket
}


proc send_string { s } {
    global socket
    send_integer [string length $s]
    puts -nonewline $socket $s
    flush $socket
}


proc receive_short {} {
    global socket
    binary scan [read $socket 2] S n
    return $n
}


proc receive_integer {} {
    global socket
    binary scan [read $socket 4] I n
    return $n
}


proc receive_string {} {
    global socket
    set len [receive_integer]
    if {$len > 0} {
	return [read $socket $len]
    } else {
	return ""
    }
}


proc rpc_validate {} {
    global socket
    if {"[string index [read $socket 1] 0]" != "\x00"} {
       error [receive_string]
    }
}


proc open { username password schema } {
    global socket server port_number
    set socket [socket $server $port_number]
    fconfigure $socket -blocking 1 -buffersize 10000 -translation {binary binary}
    command open
    send_string $username
    send_string $password
    send_string $schema
    rpc_validate
}


proc close {} {
    command close
    rpc_validate
    send_short -1
}


proc begin {} {
    command begin
    rpc_validate
}


proc commit {} {
    command commit
    rpc_validate
}


proc query { q } {
    command query
    send_string $q
    rpc_validate
    return [receive_integer]
}


proc getType { obj } {
    command getType
    send_integer $obj
    rpc_validate
    return [receive_string]
}


proc toString { obj } {
    command toString
    send_integer $obj
    rpc_validate
    return [receive_string]
}

proc tclschema {} {
    command tclschema
    rpc_validate
    return [receive_string]
}
