/*********************************************************************************
 *
 *  The LDB server
 *  Copyright (c) 1999-2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 9/30/02
 *
 ********************************************************************************/

#include <signal.h>
#include "gc_cpp.h"
#include "ldb_sm.h"
#include "sm_vas.h"
#include "odmg_main.h"

void sm_error ( const char* file, int line, rc_t rc );

rc_t init_config_options ( const char* opt_file, option_group_t& options );


void abort (int) {
  cerr << "Aborting the Server.\n";
  w_base_t::abort();
};


/* in case of RPC connection, limit the output size to 10KB */
extern bool rpc_printer_limitation;


void initialize_database () {
  ldb_sm->begin_transaction();
  create_user(sm_string("admin"),sm_string("pass"));
  ldb_sm->commit_transaction();
};


int main ( int argc, char* argv[] ) {
  option_t* volume_opt;
  option_t* volumesize_opt;
  const int option_level_cnt = 3;
  option_group_t options(option_level_cnt);
  SM_DO(options.add_option("sm_volume", "unix pathname",
			   "./database", "the database volume",
			   false, option_t::set_value_charstr,
			   volume_opt));
  SM_DO(options.add_option("sm_volumesize", "size in KBytes",
			   "1000", "the database volume size",
			   false, option_t::set_value_long,
			   volumesize_opt));
  SM_DO(options.add_class_level("ssm"));
  SM_DO(options.add_class_level("server"));
  SM_DO(options.add_class_level("ldb"));
  SM_DO(init_config_options(LDBconfig,options));
  if (argc==2 && strcmp(argv[1],"-build")==0)
  {  cout << "Do you really want to format the entire database? ";
     char answer;
     cin >> answer;
     if (answer != 'y' && answer != 'Y')
        SM_ERROR("Please try again without the -build option");
     ldb_sm = new storage_manager(sm_string(volume_opt->value()),true,
				  strtol(volumesize_opt->value(),0,0));
     ldb_sm->run(initialize_database);
  } else {
     signal(SIGINT,abort);
     signal(SIGQUIT,abort);
     signal(SIGFPE,abort);
     signal(SIGSEGV,abort);
     try {
       rpc_printer_limitation = true;
       ldb_sm = new storage_manager(sm_string(volume_opt->value()),false);
       ldb_sm_location = store_object(ldb_sm);
       ldb_sm->run();
     } catch (odmg_error_class(m)) {
       cerr << m.message;
       w_base_t::abort();
     };
  };
};
