/********************************************************************************
*
* The basic library for strings, lists, exprs etc
*
* Copyright (c) 1999-2003 by Leonidas Fegaras, the University of Texas at
*     Arlington. All rights reserved.
* Programmer: Leonidas Fegaras, UTA
* Date: 9/10/97
*
*********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <strstream.h>
#include <string.h>
#include <stdarg.h>
#include "external.h"


int string::length () const { return strlen(sp); };

string::string () { sp = NULL; };

string::string ( const char* s ) { sp = new (GC) char[strlen(s)+1]; strcpy(sp,s); };

bool string::operator == ( const string a ) const { return sp == a.sp || strcmp(sp,a.sp)==0; };

bool string::operator != ( const string a ) const { return strcmp(sp,a.sp)!=0; };

bool string::operator >= ( const string a ) const { return strcmp(sp,a.sp)>=0; };

bool string::operator <= ( const string a ) const { return strcmp(sp,a.sp)<=0; };

bool string::operator > ( const string a ) const { return strcmp(sp,a.sp)>0; };

bool string::operator < ( const string a ) const { return strcmp(sp,a.sp)<0; };

ostream& operator << ( ostream& s, const string a ) { return s << a.content(); };

char string::nth ( int n ) const { return sp[n]; };


const string string::substring ( int n, int size ) const {
  string r;
  if (n>length())
     return r;
  r.sp = new (GC) char[size];
  strncpy(r.sp,sp+n,size);
  return r;
};


int string::position ( const string s ) const {
  char* ptr;
  ptr = strstr(sp,s.sp);
  if (ptr==NULL)
     return -1;
  return (int) ptr - (int) sp;
};


const string operator+ ( const string a, const string b ) {
  string r;
  r.sp = new (GC) char[strlen(a.sp)+strlen(b.sp)+1];
  strcpy(r.sp,a.sp);
  strcat(r.sp,b.sp);
  return r;
};


const string operator+ ( const string a, const char* s) {
  string r;
  r.sp = new (GC) char[strlen(s)+strlen(a.sp)+1];
  strcpy(r.sp,a.sp);
  strcat(r.sp,s);
  return r;
};


const string operator+ ( const char* a, const string b ) {
  string r;
  r.sp = new (GC) char[strlen(a)+strlen(b.sp)+1];
  strcpy(r.sp,a);
  strcat(r.sp,b.sp);
  return r;
};


unsigned long string::hash () const {
  unsigned long res = 0;
  for(int i = strlen(sp)-1; i>=0; i--)
     res = (res<<3) | (sp[i]);
  return res;
};


String reads ( istream& from ) {
  int len;
  from >> len;
  char c;
  from.get(c);
  char* buf = new (GC) char[len+1];
  for(int i = 0; i<len; i++)
     from.get(buf[i]);
  buf[len] = '\0';
  return new string(buf);
};


void string::write ( ostream& to ) const {
  to << ' ' << strlen(sp) << ' ' << sp;
};


const char* fform ( const char* fmt, ... ) {
  int size = 100;
  char* p;
  va_list ap;
  if ((p = new (GC) char[size]) == NULL)
     return NULL;
  while (true) {
    va_start(ap, fmt);
    int n = vsnprintf(p, size, fmt, ap);
    va_end(ap);
    if (n > -1 && n < size)
      return p;
    if (n > -1)
       size = n+1;
    else size *= 2;
    delete p;
    if ((p = new (GC) char[size]) == NULL)
       return NULL;
  };
};


long long_size ( const long n ) {
  if (n==0)
     return 1;
  long size = (long) ceil(log10((double) abs(n)));
  if (n<0)
     return size+1;
  else return size;
};


string::string ( const int n ) {
  sp = new (GC) char[long_size((long)n)+1];
  sprintf(sp,"%i",n);
};


string::string ( const long n ) {
  sp = new (GC) char[long_size(n)+1];
  sprintf(sp,"%li",n);
};


string::string ( const double x ) {
  const char* s = fform("%d",x);
  sp = new (GC) char[strlen(s)+1];
  strcpy(sp,s);
};


string* new_name () {
  return new string("_X" + string(name_counter++));
};


/*--------------------------------------------------------------------------------*/


Expr variable ( String s ) {
  expr *e = new expr;
  e->tag = expr::VAR;
  e->info.Variable = new string(*s);
  return e;
};


Expr integer ( long n ) {
  expr *e = new expr;
  e->tag = expr::INT;
  e->info.Integer = n;
  return e;
};


Expr real ( double n ) {
  expr *e = new expr;
  e->tag = expr::REALN;
  e->info.Real = n;
  return e;
};


Expr estring ( String s ) {
  expr *e = new expr;
  e->tag = expr::STRING;
  e->info.Estring = new string(*s);
  return e;
};


Expr function ( Expr s, list<Expr>* r ) {
   expr *e = new expr;
   e->tag = expr::FNC;
   e->info.Function.Name = s;
   e->info.Function.Parameters = r;
   return e;
};


bool member ( Expr e, list<Expr>* r ) {
   for(list<Expr>* s = r; s->consp(); s=s->tl)
      if (s->hd->eq(e))
	 return true;
   return false;
};


bool member ( Expr e, list<String>* r ) {
   for(list<String>* s = r; s->consp(); s=s->tl)
      if (*s->hd == *e->info.Variable)
	 return true;
   return false;
};


ostream& operator<< ( ostream& s, Expr e ) {
  if (e->variablep()) return s << e->info.Variable;
  else if (e->integerp()) return s << e->info.Integer;
  else if (e->realp()) return s << e->info.Real;
  else if (e->stringp()) return s << '\"' << e->info.Estring << '\"';
  else if (e->functionp())
       return s << e->info.Function.Name << e->info.Function.Parameters;
  else return s << "** unknown Expr type **";
};


String expr::into_string () const {
  if (variablep()) return info.Variable;
  else if (integerp()) return new string(info.Integer);
  else if (realp()) return new string(info.Real);
  else if (stringp()) return new string("\"" + *info.Estring + "\"");
  else if (functionp())
       { string s = *info.Function.Name->into_string() + "(";
	 for(list<Expr>* r = info.Function.Parameters; r->consp(); r=r->tl)
	    s = s + *r->hd->into_string() + (r->tl->consp() ? "," : "");
	 return new string(s + ")");
       };
};


int expr::size () const {
  if (variablep()) return info.Variable->length();
  else if (integerp()) return long_size(info.Integer);
  else if (realp()) return string(info.Real).length();
  else if (stringp()) return info.Estring->length()+2;
  else if (functionp())
       { int n = info.Function.Name->size()+1;
	 for(list<Expr>* r = info.Function.Parameters; r->consp(); r=r->tl)
	    n += r->hd->size()+1;
	 return n;
	};
  return 0;
};


Expr expr::copy () const {
  const Expr cc = integer(0);
  if (functionp())
  {  list<Expr>* s = Cons(cc,Nil);
     list<Expr>* params = s;
     for(list<Expr>* r = info.Function.Parameters; r->consp(); r=r->tl)
     {  s->tl = s->tl->cons(r->hd->copy());
        s = s->tl;
     };
     return function(info.Function.Name->copy(),params->tl);
  };
  return (Expr) this;
};


void expr::pretty_print ( int pos, ostream& s ) const {
  if (variablep()) s << info.Variable;
  else if (integerp()) s << info.Integer;
  else if (realp()) s << info.Real;
  else if (stringp()) s << '\"' << info.Estring << '\"';
  else if (functionp())
	  if (info.Function.Parameters->consp() && pos+size()>screen_size)
	  { s << info.Function.Name << '(';
	    int new_pos = pos+info.Function.Name->size()+1;
	    info.Function.Parameters->hd->pretty_print(new_pos,s);
	    for(list<Expr>* r=info.Function.Parameters->tl; r->consp(); r=r->tl)
	    { s << ",\n";
	      for(int i=0; i<new_pos; i++) s << ' ';
	      r->hd->pretty_print(new_pos,s);
	    };
	    s << ")";
	  }
	  else s << info.Function.Name << info.Function.Parameters;
  else s << "** unknown Expr type **";
};


bool expr::eq ( Expr e ) const {
  if (this==e) return true;
  else if (e->tag==ERROR || tag==ERROR) return false;
  else if (variablep() && *info.Variable == "_") return true;
  else if (e->variablep() && *e->info.Variable == "_") return true;
  else if (variablep() && e->variablep())
       return *info.Variable == *e->info.Variable;
  else if (integerp() && e->integerp())
       return (info.Integer==e->info.Integer);
  else if (realp() && e->realp())
       return (info.Real==e->info.Real);
  else if (stringp() && e->stringp())
       return *info.Estring == *e->info.Estring;
  else if (functionp() && e->functionp())
          if (info.Function.Name->eq(e->info.Function.Name))
	     { bool b = true;
	       list<Expr>* s = e->info.Function.Parameters;
	       list<Expr>* r = info.Function.Parameters;
	       for(;b && r->consp() && s->consp(); r=r->tl, s=s->tl)
		  b = b && r->hd->eq(s->hd);
	       return (b && r->nullp() && s->nullp());
             }
          else return false;
  else return false;
};


unsigned long expr::hash () {
  unsigned long res = 0;
  if (hash_value>0) res = hash_value;
  else if (variablep()) res = info.Variable->hash();
  else if (integerp()) res = info.Integer;
  else if (realp()) res = (long) ceil(info.Real*100.0);
  else if (stringp()) res = info.Estring->hash();
  else if (functionp())
       { res = info.Function.Name->hash();
         for(list<Expr>* r = info.Function.Parameters; r->consp(); r = r->tl)
            res = res ^ (r->hd->hash());
       };
  hash_value = res;
  return res;
};


Expr subst_expr ( Expr term, Expr value, Expr e ) {
  if (e->eq(term))
     return value;
  else if (e->functionp())
       return function(subst_expr(term,value,e->info.Function.Name),
		       subst_list(term,value,e->info.Function.Parameters));
  else return e;
};


list<Expr>* subst_list ( Expr term, Expr value, list<Expr>* es ) {
  list<Expr>* res = Nil;
  for(list<Expr>* r = es; r->consp(); r=r->tl)
     res = Cons(subst_expr(term,value,r->hd),res);
  return res->reverse();
};


Expr read ( istream& from ) {
  Expr res = new expr;
  short tag;
  char c;
  from >> tag;
  from.get(c);
  res->tag = (expr::tagtype) tag;
  if (res->variablep()) res->info.Variable = reads(from);
  else if (res->integerp()) from >> res->info.Integer;
  else if (res->realp()) from >> res->info.Real;
  else if (res->stringp()) res->info.Estring = reads(from);
  else if (res->functionp())
       { int len;
	 res->info.Function.Name = read(from);
	 from >> len;
	 res->info.Function.Parameters = Nil;
	 for(int i=0; i<len; i++)
	    res->info.Function.Parameters = res->info.Function.Parameters->cons(read(from));
	 res->info.Function.Parameters = res->info.Function.Parameters->reverse();
       };
  return res;
};


void expr::write ( ostream& to ) const {
  to << ' ' << tag;
  if (variablep()) info.Variable->write(to);
  else if (integerp()) to << ' ' << info.Integer;
  else if (realp()) to << ' ' << info.Real;
  else if (stringp()) info.Estring->write(to);
  else if (functionp())
       { info.Function.Name->write(to);
	 to << ' ' << info.Function.Parameters->length();
         for(list<Expr>* r = info.Function.Parameters; r->consp(); r = r->tl)
	     r->hd->write(to);
       };
};


Expr string_to_expr ( String s ) {
  istrstream si(s->content());
  return read(si);
};


int read_word ( const char* s, char* &word ) {
  int loc = 0;
  for(; s[loc] != ' ' && s[loc] != '\0'; loc++);
  word = new (GC) char[loc+1];
  strncpy(word,s,loc);
  word[loc] = '\0';
  return loc+1;
};


int read_string ( const char* s, const int size, char* &str ) {
  str = new (GC) char[size+1];
  strncpy(str,s,size);
  str[size] = '\0';
  return size+1;
};


Expr read_expr ( const char* s, int &loc ) {
  char* word;
  loc += read_word(s+loc,word);
  short tag = atoi(word);
  if (tag == expr::FNC)
  { Expr name = read_expr(s,loc);
    loc += read_word(s+loc,word);
    int len = atoi(word);
    list<Expr>* params = Nil;
    for(int i=0; i<len; i++)
       params = params->cons(read_expr(s,loc));
    return function(name,params->reverse());
   };
   loc += read_word(s+loc,word);
   if (tag == expr::VAR)
      { loc += read_string(s+loc,atoi(word),word);
        return variable(new string(word));
      }
   else if (tag == expr::INT)
      return integer(atoi(word));
   else if (tag == expr::REALN)
      return real(atof(word));
   else if (tag == expr::STRING)
      { loc += read_string(s+loc,atoi(word),word);
        return estring(new string(word));
      }
   else return new expr();
};


Expr read_expr ( const char* s ) {
  int loc = 0;
  return read_expr(s+1,loc);
};
