/*--------------------------------------------------------------------------------
-  The template method's code.
-  Must be included here, since templates are macros
--------------------------------------------------------------------------------*/

template<class T,class S>
ostream& operator<< ( ostream& s, pair<T,S>* p ) {
  return s << p->first << ": " << p->second;
};


template<class T>
list<T>* list<T>::cons ( const T a ) {
  list<T> *n = new list<T>;
  n->tag = Cons;
  n->hd = a;
  n->tl = this;
  return n;
};


template<class T>
int list<T>::length () const {
  int n = 0;
  for(list<T>* s=(list<T>*)this; s->consp(); s=s->tl)
     n++;
  return n;
};


template<class T>
list<T>* list<T>::append ( list<T>* x ) const {
  if ( nullp() )
     return x;
  else if ( x->nullp() )
     return (list<T>*)this;
  else {
     list<T>* z = x->cons(x->hd);
     list<T>* res = z;
     for(list<T>* s=(list<T>*)this; s->consp(); s=s->tl)
     {  z->tl = z->tl->cons(s->hd);
        z = z->tl;
     };
     return res->tl;
  };
};


template<class T>
list<T>* list<T>::reverse () const {
  list<T>* res = new list<T>;
  for(list<T>* s=(list<T>*)this; s->consp(); s=s->tl)
     res = res->cons(s->hd);
  return res;
};


/* the nth element of a list */
template<class T>
const T list<T>::nth ( int n ) const {
  list<T>* r = (list<T>*)this;
  for(int i = 1; r->consp() && i<n; i++, r=r->tl) ;
  if (r->consp())
     return r->hd;
};


/* returns true if x is a subset of y */
template<class T>
bool list<T>::subset ( list<T>* x, bool (eq)(const T,const T) ) const {
  for(list<T>* r = (list<T>*)this; r->consp(); r=r->tl)
     if (!x->member(r->hd,eq))
        return false;
  return true;
};


template<class T>
bool list<T>::member ( T item, bool (eq)(const T,const T) ) const {
  for(list<T>* r = (list<T>*)this; r->consp(); r=r->tl)
     if (eq(r->hd,item))
	return true;
  return false;
};


template<class T>
list<T>* list<T>::sort ( bool (less_than) (const T,const T) ) const {
  int len = length();
  T A[len];
  int k = 0;
  for(list<T>* r = (list<T>*)this; r->consp(); r=r->tl)
     A[k++] = r->hd;
  for(int j = len-1; j>=1; j--)
     for(int i = 1; i<=j; i++)
	if (less_than(A[i],A[i-1]))
	{ T tmp = A[i-1];
	  A[i-1] = A[i];
	  A[i] = tmp;
	};
  list<T>* res = new list<T>;
  for(int i = len-1; i>=0; i--)
     res = res->cons(A[i]);
  return res;
};


template<class T>
ostream& operator<<( ostream& s, list<T> *lst ) {
  if (lst->nullp())
     return s << "()";
  s << "(" << lst->hd;
  for( list<T>* r = lst->tl; r->consp(); r = r->tl)
     s << "," << r->hd;
  return s << ")";
};


template<class T>
binding<T>::binding ( String s, const T c ) {
  env = (new list<pair<String,T>*>)->cons(new pair<String,T>(s,c));
};


template<class T>
binding<T>* binding<T>::extend ( pair<String,T>* x ) const {
  binding<T>* res = new binding<T>(this);
  res->env = res->env->cons(x);
  return res;
};


template<class T>
binding<T>* binding<T>::extend ( binding<T>* r ) const {
  binding<T>* res = new binding<T>(this);
  res->env = res->env->append(r->env);
  return res;
};


template<class T>
const T binding<T>::find ( String n ) const {
  for(list<pair<String,T>*>* r = env; r->consp(); r=r->tl)
     if (r->hd->first->eq(n)) return r->hd->second;
};


template<class T>
const T binding<T>::find ( String n, const T def ) const {
  for(list<pair<String,T>*>* r = env; r->consp(); r=r->tl)
     if (r->hd->first->eq(n)) return r->hd->second;
  return def;
};


template<class T>
Names binding<T>::names () const {
  list<String>* s = new list<String>;
  for(list<pair<String,T>*>* r = env; r->consp(); r=r->tl)
     s = s->cons(r->hd->first);
  return s->reverse();
};


template<class T>
bool binding<T>::in ( String n ) const {
  for(list<pair<String,T>*>* r = env; r->consp(); r=r->tl)
     if (r->hd->first->eq(n)) return true;
  return false;
};


template<class T>
ostream& operator<<( ostream& s, binding<T> *lst ) {
  if (lst->env->nullp())
     return s << "[]";
  s << "[" << lst->env->hd;
  for( list<pair<String,T>*>* r = lst->env->tl; r->consp(); r = r->tl)
     s << "," << r->hd;
  return s << "]";
};
