/********************************************************************************
*
* Some constant OQL methods
*
*  Copyright (c) 1999-2003 by Leonidas Fegaras, the University of Texas at
*     Arlington, All rights reserved.
* Programmer: Leonidas Fegaras, UTA
* Date: 4/26/97
*
********************************************************************************/

#include "external.h"
#include "odl.h"
#include "eval.h"
#include <math.h>


Expr convert ( const char* s ) {
  if (strcmp(s,"int")==0)
     return variable("integer");
  else if (strcmp(s,"bool")==0)
     return variable("boolean");
  else if (strcmp(s,"sm_string")==0)
     return variable("string");
  else if (strcmp(s,"float")==0)
     return variable("float");
  else if (strcmp(s,"oid")==0)
     return variable("oid");
};


typedef struct { String         name;
                 list<Expr>*    parameters;
                 int            location;
               } method_info;


method_info* make_method_info ( String name, list<Expr>* parameters, int location ) {
   method_info* m = new (GC) method_info;
   m->name = name;
   m->parameters = parameters;
   m->location = location;
   return m;
};


binding<method_info*>* internal_methods = new binding<method_info*>;

#define SF(name,params,fncs) store_function(#name,(void*) &name,params,fncs)
#define T2(name,opr,ot,xt,yt)                                                   \
        void* name##_##xt##_##yt ( void* x, void* y ) {                         \
           return (void*) new (GC) ot(*((xt*) x) opr *((yt*) y));               \
        };                                                                      \
        int name##_##xt##_##yt##_ =                                             \
        (internal_methods = internal_methods->extend(                           \
                           new string(#name),                                   \
                           make_method_info(new string(#opr),                   \
                                Nil->cons(convert(#yt))                         \
                                   ->cons(convert(#xt))                         \
                                   ->cons(convert(#ot)),                        \
                                SF(name##_##xt##_##yt,2,0))),0);
#define T2B(name,opr,ot,xt,yt)                                                  \
        void* name##_##xt##_##yt ( void* x, void* y ) {                         \
           return (void*) new (GC) ot(opr(*((xt*) x),*((yt*) y)));              \
        };                                                                      \
        int name##_##xt##_##yt##_ =                                             \
                (internal_methods = internal_methods->extend(                   \
                           new string(#name),                                   \
                           make_method_info(new string(#opr),                   \
                                Nil->cons(convert(#yt))                         \
                                   ->cons(convert(#xt))                         \
                                   ->cons(convert(#ot)),                        \
                                SF(name##_##xt##_##yt,2,0))),0);
#define T1B(name,opr,ot,xt)                                                     \
        void* name##_##xt ( void* x ) {                                         \
           return (void*) new (GC) ot(opr(*((xt*) x)));                         \
        };                                                                      \
        int name##_##xt##_ =                                                    \
                (internal_methods = internal_methods->extend(                   \
                           new string(#name),                                   \
                           make_method_info(new string(#opr),                   \
                                Nil->cons(convert(#xt))                         \
                                   ->cons(convert(#ot)),                        \
                                SF(name##_##xt,1,0))),0);


typedef sm_ref oid;


int string_compare ( sm_string x, sm_string y ) { return x.strcmp(y); };
bool string_eq ( sm_string x, sm_string y ) { return string_compare(x,y) == 0; };
bool string_neq ( sm_string x, sm_string y ) { return string_compare(x,y) != 0; };
bool string_gt ( sm_string x, sm_string y ) { return string_compare(x,y) > 0; };
bool string_lt ( sm_string x, sm_string y ) { return string_compare(x,y) < 0; };
bool string_geq ( sm_string x, sm_string y ) { return string_compare(x,y) >= 0; };
bool string_leq ( sm_string x, sm_string y ) { return string_compare(x,y) <= 0; };
float abs ( float x ) { return (x>=0) ? x : -x; };
int abs ( int x ) { return (x>=0) ? x : -x; };
int mod ( int x, int y ) { return x % y; };


//         name     operation          output  left    right
//----------------------------------------------------------------------
        T2( plus,        +,              int,    int,    int     );
        T2( plus,        +,              float , float , int     );
        T2( plus,        +,              float , int,    float   );
        T2( plus,        +,              float , float , float   );
        T2( minus,       -,              int,    int,    int     );
        T2( minus,       -,              float , float , int     );
        T2( minus,       -,              float , int,    float   );
        T2( minus,       -,              float , float , float   );
        T2( times,       *,              int,    int,    int     );
        T2( times,       *,              float , float , int     );
        T2( times,       *,              float , int,    float   );
        T2( times,       *,              float , float , float   );
        T2( div,         /,              int,    int,    int     );
        T2( div,         /,              float , float , int     );
        T2( div,         /,              float , int,    float   );
        T2( div,         /,              float , float , float   );
        T2B( mod,        mod,            int,    int,    int     );
        T1B( abs,        abs,            int,    int             );
        T1B( minus,      -,              int,    int             );
        T1B( abs,        abs,            float , float           );
        T1B( minus,      -,              float , float           );
        T2( eq,          ==,             bool,   bool,   bool    );
        T2( neq,         !=,             bool,   bool,   bool    );
        T2( and,         &&,             bool,   bool,   bool    );
        T2( or,          ||,             bool,   bool,   bool    );
        T1B( not,        !,              bool,   bool            );
        T2( eq,          ==,             bool,   int,    int     );
        T2( neq,         !=,             bool,   int,    int     );
        T2( gt,          >,              bool,   int,    int     );
        T2( lt,          <,              bool,   int,    int     );
        T2( geq,         >=,             bool,   int,    int     );
        T2( leq,         <=,             bool,   int,    int     );
        T2( eq,          ==,             bool,   float , float   );
        T2( neq,         !=,             bool,   float , float   );
        T2( gt,          >,              bool,   float , float   );
        T2( lt,          <,              bool,   float , float   );
        T2( geq,         >=,             bool,   float , float   );
        T2( leq,         <=,             bool,   float , float   );
        T2( gt,          >,              bool,   float , int     );
        T2( lt,          <,              bool,   float , int     );
        T2( geq,         >=,             bool,   float , int     );
        T2( leq,         <=,             bool,   float , int     );
        T2( gt,          >,              bool,   int,    float   );
        T2( lt,          <,              bool,   int,    float   );
        T2( geq,         >=,             bool,   int,    float   );
        T2( leq,         <=,             bool,   int,    float   );
        T2B( eq,         string_eq,      bool,   sm_string,     sm_string      );
        T2B( neq,        string_neq,     bool,   sm_string,     sm_string      );
        T2B( gt,         string_gt,      bool,   sm_string,     sm_string      );
        T2B( lt,         string_lt,      bool,   sm_string,     sm_string      );
        T2B( geq,        string_geq,     bool,   sm_string,     sm_string      );
        T2B( leq,        string_leq,     bool,   sm_string,     sm_string      );
        T2B( plus,       string_concat,  sm_string,     sm_string,     sm_string      );
