#include "external.h"
#include <stdlib.h>
#include <stream.h>
#include <sys/socket.h>
#include <string.h>
#include "rpc.h"


void ldb_abort ( const char* m );

void rpc_error ( const char* msg ) {
#ifdef LDB_SERVER
  eout << "RPC Error: " << msg << ".\n";
  ldb_abort(eout.str());
#else
  cerr << "RPC Error: " << msg << ".\n";
  RPC_stream << (short) -1;
  exit(-1);
#endif
};


rpc_stream& rpc_stream::rpc_send ( void* data, size_t size ) {
  if (send(sockfd,(char*)data,size,0) == size)
     return *this;
  else rpc_error("Incomplete data in socket");
};


rpc_stream& rpc_stream::rpc_receive ( void* &data, size_t size ) {
  if (recv(sockfd,(char*)data,size,MSG_WAITALL) == size)
     return *this;
  else rpc_error("Incomplete data from socket");
};


void rpc_stream::rpc_OK () {
  char res = 0;
  send(sockfd,&res,1,0);
};


void rpc_stream::rpc_validate () {
  char res;
  recv(sockfd,&res,1,MSG_WAITALL);
  if (res!=0)
     cerr << "got it!\n";
};


rpc_stream& operator<< ( rpc_stream& s, const char* v ) {
  int len = strlen(v);
  s << len;
  return s.rpc_send((void*) v,len); 
};


rpc_stream& operator>> ( rpc_stream& s, char* &v ) {
  int len;
  s >> len;
  v = new (GC) char[len+1];
  v[len] = '\0';
  return s.rpc_receive((void*) v,len);
};


rpc_stream RPC_stream;

int RPC_fnc_count = 0;

dispatch_table_t dispatch_table[100];

const int max_number_of_objects = 1000;

static int number_of_objects = 0;

static void* all_objects[max_number_of_objects];


int store_object ( void* obj ) {
  if (number_of_objects>=max_number_of_objects)
     rpc_error("Cannot store more objects on the server");
  all_objects[number_of_objects] = obj;
  return number_of_objects++;
};


void* get_object ( int n ) {
  if (n<0 || n>=number_of_objects)
     rpc_error(fform("Wrong RPC object reference %d",n));
  else return all_objects[n];
};
