#include <sys/socket.h>

class rpc_stream {
public:
  int sockfd;
  rpc_stream ( int fd ): sockfd(fd) {};
  rpc_stream () {};
  rpc_stream& rpc_send ( void* data, size_t size );
  rpc_stream& rpc_receive ( void* &data, size_t size );
  void rpc_validate ();
  void rpc_OK ();
};


rpc_stream& operator<< ( rpc_stream&, const char* );

rpc_stream& operator>> ( rpc_stream&, char*& );


#if defined(linux) || defined(__linux__)
// Need to convert big-endian data (low byte first) into little-endian
// (high byte first), since Java's DataStreams are little-endian
template<class T>
rpc_stream& operator<< ( rpc_stream& s, T v ) {
  for(int i=sizeof(T)-1; i>=0; i--)
     s.rpc_send((void*)(((char*)&v)+i),1); 
  return s;
};
template<class T>
rpc_stream& operator>> ( rpc_stream& s, T &v ) {
  for(int i=sizeof(T)-1; i>=0; i--)
  {  void* data = (void*)(((char*)&v)+i);
     s.rpc_receive(data,1);
  };
  return s;
};
#else
// little-endian (eg, DEC)
template<class T>
rpc_stream& operator<< ( rpc_stream& s, T v ) {
   return s.rpc_send(&v,sizeof(T));
};
template<class T>
rpc_stream& operator>> ( rpc_stream& s, T &v ) {
   void* data = (void*) &v;
   return s.rpc_receive(data,sizeof(T));
};
#endif


typedef void (*dispatch_table_t)(rpc_stream);


#ifdef LDB_SERVER
extern dispatch_table_t dispatch_table[100];
#endif

extern int RPC_fnc_count;

extern rpc_stream RPC_stream;

int store_object ( void* obj );

void* get_object ( int n );
